import re
from urllib.parse import urlparse, parse_qs

from lxml.objectify import fromstring
from requests.exceptions import HTTPError
import responses

from .base import OscTest, CallbackFactory
from ..extensions.projects import TEMPLATE_META


class TestProject(OscTest):
    @responses.activate
    def test_get_list(self):
        def callback(headers, params, request):
            status = 200
            body = """<directory>
                <entry name="Devel:AAC"/>
                <entry name="Devel:ARM"/>
                <entry name="Devel:ARM:Factory"/>
                <entry name="Devel:ARM:Factory:Contrib:ThunderX"/>
                <entry name="Devel:ARM:Factory:Contrib:ThunderX:qemu"/>
                <entry name="Devel:ARM:Factory:ILP32"/>
            </directory>
            """
            parsed = urlparse(request.url)
            params.update(parse_qs(parsed.query))
            if params.get("deleted", ["0"]) == ["1"]:
                status = 403
                body = """<status code="no_permission_for_deleted">
                  <summary>only admins can see deleted projects</summary>
                </status>
                """
            headers['request-id'] = '728d329e-0e86-11e4-a748-0c84dc037c13'
            return status, headers, body

        self.mock_request(
            method=responses.GET,
            url=self.osc.url + '/source',
            callback=CallbackFactory(callback)
        )

        with self.subTest("existing files"):
            response = self.osc.projects.get_list(deleted=False)
            self.assertEqual(response.tag, "directory")
            self.assertEqual(response.countchildren(), 6)

        with self.subTest("deleted files"):
            self.assertRaises(
                HTTPError, self.osc.projects.get_list, deleted=True
            )

    @responses.activate
    def test_get_meta(self):
        def callback(headers, params, request):
            status = 200
            body = """
                <project name="Devel:ARM:Factory">
                  <title>openSUSE Factory built on native ARM Hardware</title>
                  <description>openSUSE Factory built on native ARM Hardware. 
                  It usually builds faster than the external version 
                  openSUSE:Factory:ARM, which builds on qemu in x86_64.
                </description>
                  <link project="openSUSE.org:openSUSE:Factory:ARM"/>
                  <person userid="Foo Bar" role="maintainer"/>
                  <person userid="Hello World" role="maintainer"/>
                  <build>
                    <disable/>
                    <enable arch="aarch64" repository="standard"/>
                  </build>
                  <debuginfo>
                    <enable/>
                  </debuginfo>
                  <repository name="standard" rebuild="local" block="never" 
                      linkedbuild="all">
                    <path project="openSUSE.org:openSUSE:Factory:ARM" 
                      repository="standard"/>
                    <arch>armv7l</arch>
                    <arch>aarch64</arch>
                  </repository>
                  <repository name="images" block="local">
                    <path project="Devel:ARM:Factory" repository="standard"/>
                    <arch>local</arch>
                    <arch>armv7l</arch>
                  </repository>
                </project>
            """
            if not "Devel:ARM:Factory" in request.url:
                status = 404
                body = """
                    <status code="unknown_project">
                      <summary>Devel:ARM:Fbctory</summary>
                    </status>
                """
            headers['request-id'] = '728d329e-0e86-11e4-a748-0c84dc037c13'
            return status, headers, body

        self.mock_request(
            method=responses.GET,
            url=re.compile(self.osc.url + '/source/.*'),
            callback=CallbackFactory(callback)
        )

        with self.subTest("existing project"):
            response = self.osc.projects.get_meta("Devel:ARM:Factory")
            self.assertEqual(response.tag, "project")
            self.assertEqual(response.get("name"), "Devel:ARM:Factory")
            self.assertEqual(len(response.xpath("./person")), 2)

        with self.subTest("non-existing project"):
            self.assertRaises(
                HTTPError, self.osc.projects.get_meta, "Devel:ARM:Fbctory"
            )

    @responses.activate
    def test_put_meta(self):
        def callback(headers, params, request):
            status = 500
            body = "<status code='error'></status>"
            try:
                meta = fromstring(request.body)
            except Exception:
                pass
            else:
                match = re.search("source/(?P<project>[^/]+)/_meta",
                                  request.url)
                if match and meta.get("name") == match.group("project"):
                    status = 200
                    body = """<status code='ok'></status>"""

            headers['request-id'] = '728d329e-0e86-11e4-a748-0c84dc037c13'
            return status, headers, body

        self.mock_request(
            method=responses.PUT,
            url=re.compile(self.osc.url + '/source/[^/]+/_meta'),
            callback=CallbackFactory(callback)
        )

        with self.subTest("Valid request"):
            self.assertTrue(self.osc.projects.put_meta(
                project="test",
                title="foo bar",
                description="lorem ipsum dolor sit amet ..."
            ))

        with self.subTest("Invalid request"):
            self.assertRaises(
                HTTPError,
                self.osc.request,
                url=self.osc.url + '/source/foobar/_meta',
                method="PUT",
                data=TEMPLATE_META
            )

    @responses.activate
    def test_get_files(self):
        def callback(headers, params, request):
            status = 200
            body = """
                <directory name="_project" rev="10" vrev="" 
                  srcmd5="4adff519b4e14cf206bcc7bdce40a73c">
                  <entry name="_config" md5="11cfe7fdb2591b55435b312c156e17fe" 
                    size="1818" mtime="1542617696" />
                </directory>
            """
            parsed = urlparse(request.url)
            params.update(parse_qs(parsed.query))

            if params.get("meta", ['0']) == ['1']:
                status = 200
                body = """
                    <directory name="_project" rev="41" vrev="" 
                      srcmd5="f8bf4cb0e1edea467e34462742ca82ae">
                      <entry name="_attribute" 
                        md5="18fbe34f9ab146c9d73f7e5756f74d48" size="4172" 
                        mtime="1543406498" />
                      <entry name="_meta" md5="9392aa5e748eb15087c4db55ac1cb7f3"
                        size="1875" mtime="1542635759" />
                    </directory>
                """
            headers['request-id'] = '728d329e-0e86-11e4-a748-0c84dc037c13'
            return status, headers, body

        self.mock_request(
            method=responses.GET,
            url=re.compile(
                self.osc.url + '/source/SUSE:SLE-15-SP1:GA/_project.*'
            ),
            callback=CallbackFactory(callback)
        )

        with self.subTest("Non-meta files"):
            response = self.osc.projects.get_files("SUSE:SLE-15-SP1:GA",
                                                   "_project")
            self.assertEqual(response.tag, "directory")
            self.assertEqual(len(response.xpath("./entry")), 1)

        with self.subTest("Meta files"):
            response = self.osc.projects.get_files(
                "SUSE:SLE-15-SP1:GA", "_project", meta=True
            )
            self.assertEqual(response.tag, "directory")
            self.assertEqual(len(response.xpath("./entry")), 2)

    @responses.activate
    def test_get_attribute(self):
        def callback(headers, params, request):
            status = 404
            body = """
            <status code="unknown_attribute_type">
              <summary>
                Attribute Type OBS:ApprovedRequestSour does not exist
              </summary>
            </status>
            """
            if request.url.endswith("_attribute"):
                status = 200
                body = """
                <attributes>
                  <attribute name="IgnoredIssues" namespace="OSRT">
                    <value>last_seen: {}</value>
                  </attribute>
                  <attribute name="OpenQAMapping" namespace="OSRT">
                    <value>s/^(SLE-.*)-Installer-(.*)Build/\g&lt;1&gt;
                    -Staging:$LETTER-Installer-\g&lt;2&gt;Build$LETTER./</value>
                  </attribute>
                  <attribute name="ApprovedRequestSource" namespace="OBS"/>
                </attributes>
                """
            if request.url.endswith("OBS:ApprovedRequestSource"):
                status = 200
                body = """
                    <attributes>
                      <attribute name="ApprovedRequestSource" namespace="OBS"/>
                    </attributes>
                """
            headers['request-id'] = '728d329e-0e86-11e4-a748-0c84dc037c13'
            return status, headers, body

        self.mock_request(
            method=responses.GET,
            url=re.compile(
                self.osc.url + '/source/SUSE:SLE-15-SP1:GA/_attribute.*'
            ),
            callback=CallbackFactory(callback)
        )

        with self.subTest("all attributes"):
            response = self.osc.projects.get_attribute("SUSE:SLE-15-SP1:GA")
            self.assertEqual(response.tag, "attributes")
            self.assertEqual(len(response.xpath("./attribute")), 3)

        with self.subTest("one attribute"):
            response = self.osc.projects.get_attribute(
                "SUSE:SLE-15-SP1:GA", "OBS:ApprovedRequestSource"
            )
            self.assertEqual(response.tag, "attributes")
            self.assertEqual(len(response.xpath("./attribute")), 1)

        with self.subTest("wrong attribute"):
            self.assertRaises(
                HTTPError, self.osc.projects.get_attribute,
                "SUSE:SLE-15-SP1:GA", "FOO:Bar"
            )

    @responses.activate
    def test_set_attribute(self):
        def callback(headers, params, request):
            status, body = 200, "<status code='ok'></status>"
            return status, headers, body

        self.mock_request(
            method="POST",
            url=re.compile(
                self.osc.url + '/source/(?P<project>)[^/]+/_attribute/?'
            ),
            callback=CallbackFactory(callback)
        )

        self.assertTrue(
            self.osc.projects.set_attribute(
                project="test:project",
                attribute="namespace:attr",
                value="value"
            )
        )

    @responses.activate
    def test_delete_attribute(self):
        error = False

        def callback(headers, params, request):
            if error:
                status, body = 404, ""
            else:
                status, body = 200, "<status code='ok'></status>"
            return status, headers, body

        self.mock_request(
            method="DELETE",
            url=re.compile(
                self.osc.url + '/source/(?P<project>)[^/]+/_attribute/?'
            ),
            callback=CallbackFactory(callback)
        )

        with self.subTest("existing attr"):
            self.assertTrue(
                self.osc.projects.delete_attribute(
                    project="test:project",
                    attribute="namespace:attr"
                )
            )

        with self.subTest("non-existent attr"):
            error = True
            self.assertRaises(
                HTTPError,
                self.osc.projects.delete_attribute,
                project="test:project",
                attribute="namespace:attr"
            )

    @responses.activate
    def test_delete(self):
        def callback(headers, params, request):
            if "existing" in request.url:
                status, body = 200, """
                <status code="ok">
                  <summary>Ok</summary>
                </status>
                """
            else:
                status, body = 400, """
                    <status code="invalid_project_name">
                      <summary>
                        invalid project name 'home:apritschet:__'
                      </summary>
                    </status>"""
            return status, headers, body

        self.mock_request(
            method="DELETE",
            url=re.compile(
                self.osc.url + '/source/(?P<project>)[^/]+'
            ),
            callback=CallbackFactory(callback)
        )

        with self.subTest("existing"):
            self.assertTrue(self.osc.projects.delete("test_project:existing"))

        with self.subTest("non-existent"):
            self.assertRaises(
                HTTPError,
                self.osc.projects.delete,
                "test:project:non-found"
            )

    @responses.activate
    def test_exists(self):
        def callback(headers, params, request):
            if "existing" in request.url:
                status, body = 200, ""
            else:
                status, body = 404, ""
            return status, headers, body

        self.mock_request(
            method="HEAD",
            url=re.compile(
                self.osc.url + '/source/(?P<project>)[^/]+'
            ),
            callback=CallbackFactory(callback)
        )

        with self.subTest("existing"):
            self.assertTrue(self.osc.projects.exists("home:user:existing"))

        with self.subTest("non-existent"):
            self.assertFalse(self.osc.projects.exists("no:such:project"))



"""
Build result extension
----------------------
"""
# pylint: disable=too-few-public-methods
from __future__ import unicode_literals
from six.moves.urllib.parse import urljoin

from ..utils.base import ExtensionBase


class Build(ExtensionBase):
    """
    Osc extension for retrieving build results
    """
    base_path = "/build"

    def get(self, project, package=None, repo=None, **params):
        """
        Get build infos

        If no parameters are supplied, a compact list of build states for all
        packages in the project are returned.

        If a ``package`` and/or ``repo`` are supplied, an extended list of build
        results for these particular values is returned.

        :param project: Project name
        :param package: Package name
        :param repo: Repository name
        :param params: Additional parameters
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement

        .. versionadded:: 0.1.1
        """
        params["package"] = package
        params["repository"] = repo
        response = self.osc.request(
            method="GET",
            url=urljoin(self.osc.url, "{}/{}/_result".format(self.base_path,
                                                             project)),
            params=params
        )

        return self.osc.get_objectified_xml(response)

    def get_history(self, project, package, repo, arch="x86_64", **params):
        """
        Get build history

        :param project: Project name
        :param package: Package name
        :param repo: Repository name
        :param arch: Architecture name
        :param params: Additional parameters
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement

        .. versionadded:: 0.5.0
        """

        response = self.osc.request(
            method="GET",
            url=urljoin(self.osc.url, "{}/{}/{}/{}/{}/_history".format(self.base_path,
                                                             project,repo,arch,package)),
            params=params
        )

        return self.osc.get_objectified_xml(response)

    def get_package_list(self, project, repo, arch):
        """
        Get a list of packages for which build results exist

        :param project: Project name
        :param repo: Repository name
        :param arch: Architecture name
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement

        .. versionadded:: 0.2.4
        """
        response = self.osc.request(
            method="GET",
            url=urljoin(self.osc.url, "{}/{}/{}/{}".format(
                self.base_path, project, repo, arch
            ))
        )

        return self.osc.get_objectified_xml(response)

    def get_binary_list(self, project, repo, arch, package, **params):
        """
        Get a list of built RPMs

        :param project: Project name
        :param repo: Repository name
        :param arch: Architecture name
        :param package: Package name
        :param params: Additional parameters
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement

        .. versionadded:: 0.1.3
        """
        response = self.osc.request(
            method="GET",
            url=urljoin(self.osc.url, "{}/{}/{}/{}/{}".format(
                self.base_path, project, repo, arch, package
            )),
            params=params
        )

        return self.osc.get_objectified_xml(response)

    # pylint: disable=too-many-arguments
    def get_binary(self, project, repo, arch, package, filename):
        """
        Get the build binary file

        :param project: Project name
        :param repo: Repository name
        :param arch: Architecture name
        :param package: Package name
        :param filename: File name
        :return: Raw response
        :rtype: str

        .. versionadded:: 0.2.4
        """
        response = self.osc.request(
            method="GET",
            url=urljoin(self.osc.url, "{}/{}/{}/{}/{}/{}".format(
                self.base_path, project, repo, arch, package, filename
            )),
        )

        return response.text



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages


def get_requires():
    def _filter(requires):
        return [req.strip() for req in requires if req.strip()]

    filename = "requirements.txt"

    with open(filename, "r") as fh:
        return _filter(fh.readlines())


with open("README.md") as fh:
    long_description = fh.read()


setup(
    name='osc-tiny',
    version='0.6.1',
    description='Client API for openSUSE BuildService',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Andreas Hasenkopf',
    author_email='ahasenkopf@suse.com',
    url='http://github.com/crazyscientist/osc-tiny',
    download_url='http://github.com/crazyscientist/osc-tiny/tarball/master',
    packages=find_packages(),
    license='MIT',
    install_requires=get_requires(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ]
)



"""
Configuration utilities
^^^^^^^^^^^^^^^^^^^^^^^

This module provides a collection of utilities to access the configuration of
`osc <https://github.com/openSUSE/osc>`_ in order to make it easier to create command line tools
with OSC Tiny.

.. versionadded:: 0.4.0
"""
import typing
from base64 import b64decode
from bz2 import decompress
from configparser import ConfigParser, NoSectionError
import os
from pathlib import Path
import warnings

try:
    from osc import conf as _conf
    from osc.oscerr import ConfigError, ConfigMissingApiurl
except ImportError:
    _conf = None


def get_config_path() -> Path:
    """
    Return path of ``osc`` configuration file

    :return: Path
    :raises FileNotFoundError: if no config file found
    """
    env_path = os.environ.get("OSC_CONFIG", None)
    conf_path = os.environ.get('XDG_CONFIG_HOME', '~/.config')
    if env_path:
        path = Path(env_path)
        if path.is_file():
            return path

    for path in (Path.home().joinpath(".oscrc"),
                 Path(conf_path).joinpath("osc/oscrc").expanduser()):
        if path.is_file():
            return path

    raise FileNotFoundError("No `osc` configuration file found")


# pylint: disable=too-many-branches
def get_credentials(url: typing.Optional[str] = None) \
        -> typing.Tuple[str, str, typing.Optional[Path]]:
    """
    Get credentials for Build Service instance identified by ``url``

    .. important::

        If the ``osc`` package is not installed, this function will only try to extract the username
        and password from the configuration file.

        Any credentials stored on a keyring will not be accessible!

    :param str url: URL of Build Service instance (including schema). If not specified, the value
                    from the ``apiurl`` parameter in the config file will be used.
    :return: (username, password, SSH private key path)
    :raises ValueError: if config provides no credentials
    """
    if _conf is not None:
        try:
            _conf.get_config()
            if url is None:
                # get the default api url from osc's config
                url = _conf.config["apiurl"]
            # and now fetch the options for that particular url
            api_config = _conf.get_apiurl_api_host_options(url)
            username = api_config["user"]
            password = api_config["pass"]
            sshkey = Path(api_config["sshkey"]) if api_config["sshkey"] else None
        except (ConfigError, ConfigMissingApiurl) as error:
            if isinstance(error, ConfigError):
                raise ValueError("`osc` config was not found.") from error
            # this is the case of ConfigMissingApiurl
            raise ValueError("`osc` config has no options for URL {}".format(url)) from error

        if not username:
            raise ValueError("`osc` config provides no username for URL {}".format(url))
        if not password:
            raise ValueError("`osc` config provides no password for URL {}".format(url))
        return username, password, sshkey

    warnings.warn("`osc` is not installed. Not all configuration backends of `osc` will be "
                  "available.")
    parser = ConfigParser()
    path = get_config_path()
    parser.read((path))
    try:
        if url is None:
            url = parser["general"].get("apiurl", url)
    except (KeyError, NoSectionError) as error:
        raise ValueError("`osc` config does not provide the default API URL") from error

    if url not in parser.sections():
        raise ValueError("`osc` config has no section for URL {}".format(url))

    username = parser[url].get("user", None)
    if not username:
        raise ValueError("`osc` config provides no username for URL {}".format(url))

    password = parser[url].get("pass", None)
    if not password:
        password = parser[url].get("passx", None)
        if password:
            password = decompress(b64decode(password.encode("ascii"))).decode("ascii")

    if not password:
        raise ValueError("`osc` config provides no password for URL {}".format(url))

    sshkey = parser[url].get("sshkey", None)
    if sshkey:
        sshkey = Path(sshkey)

    return username, password, sshkey



"""
Origin extension
----------------

This is a runtime optimized implementation of the `Origin Manager`_.

The intention of this implementation is to quickly yield batch results. In doing so a different
logic than in the original OSC plugin is used:

* If a project inherits a package from another project, the origin is checked for the package in the
  original project.
* If the package has received submissions from potential origin projects, only those projects are
  considered.
* The first project from the origin configuration (including above limitations) containing the
  package is considered as origin project.
* If a project was found and if that project inherits the package, the original project is returned.

.. _Origin Manager:
    https://github.com/openSUSE/openSUSE-release-tools/blob/master/docs/origin-manager.md

.. versionadded:: 0.3.0
"""
# pylint: disable=too-many-ancestors,ungrouped-imports
from collections import defaultdict
import re
from warnings import warn

from yaml import load

from ..utils.backports import lru_cache
from ..utils.base import ExtensionBase
from ..utils.mapping import LazyOscMappable

try:
    from functools import cached_property
except ImportError:
    # Support for Python3 prior 3.8
    try:
        from backports.cached_property import cached_property
    except ImportError:
        from cached_property import cached_property

try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    # If libyaml installed, we should fall back to the SafeLoader
    from yaml import SafeLoader


class DevelProjects(LazyOscMappable):
    """
    Lazy lookup table for package specific development projects
    """

    def get_devel_projects(self, *projects):
        """
        Find package specific development projects for (maintained) projects

        :param str projects: One or more project names
        :return: Dictionary with development projects per package
        :rtype: dict
        """
        xpath = "state/@name='accepted' and action/@type='change_devel' and ({projects})"
        xpath_projects = ["action/target/@project='{}'".format(project) for project in projects]
        change_devels = self.osc.search.request(
            xpath=xpath.format(projects=" or ".join(xpath_projects))
        )

        data = {}
        for request in sorted(getattr(change_devels, "request", []),
                              key=lambda x: x.state.get("when"), reverse=True):
            key = (request.action.target.get("project"), request.action.target.get("package"))
            if key not in data:
                data[key] = request.action.source.get("project")
                break

        return data

    def __missing__(self, key):
        """
        JIT processing of missing dictionary items

        This method respects the default behavior of dictionary classes.
        """
        self._data[key] = self.get_devel_projects(key)


class PackageMatrix(LazyOscMappable):
    """
    Lazy lookup matrix for package origins

    This matrix collects for every project (provided as key) a dictionary with all allowed origin
    projects and packages therein. If the origin project inherits the package from another project,
    this information is available, too.

    The matrix is not populated at initialization time. When the matrix for a project is requested
    for the first, only the corresponding items are populated. Subsequent access to the same project
    does not trigger new population. Hence "lazy".

    Example:

    .. code-block:: python

        {
            'openSUSE:Leap:15.2:Update': {
                'SUSE:SLE-15-SP2:Update': {
                    # ...
                    'zeromq': ['SUSE:SLE-15:Update'],
                    'zypper': ['SUSE:SLE-15-SP2:Update', 'SUSE:SLE-15-SP2:GA']
                },
                'SUSE:SLE-15-SP1:Update': {
                    'zinnia-tomoe': ['SUSE:SLE-15:GA']
                    # ...
                },
                'openSUSE:Factory': {
                    # ...
                    'zziplib': ['openSUSE:Factory']
                }
            },
            'openSUSE:Leap:15.2': ...
        }

    :param osc_obj: An Osc instance to run queries
    :type osc_obj: :py:class:`osctiny.osc.Osc`
    :param expanded_origins: Lookup dictionary with all possible origin projects
    :type expanded_origins: :py:meth:`Origin.expanded_origins`
    """
    def __init__(self, osc_obj, expanded_origins, **kwargs):
        super().__init__(osc_obj=osc_obj, **kwargs)
        self._expanded_origins = expanded_origins

    def __missing__(self, key):
        """
        JIT processing of missing matrix items

        This method respects the default behavior of dictionary classes.
        """
        def _inherit(proj, o_proj):
            if proj == o_proj:
                return [proj]
            return [proj, o_proj]

        if key in self._expanded_origins:
            self._data[key] = {
                origin_project: {
                    entry.get("name"): _inherit(origin_project, entry.get("originproject"))
                    for entry in getattr(self.osc.projects.get_files(project=origin_project,
                                                                     expand='1'), "entry", [])
                }
                for origin_project in self._expanded_origins[key]
                if origin_project not in ["<devel>", "*", "*~"]
            }
            return
        raise KeyError("'{}' is not an origin project".format(key))


class Origin(ExtensionBase):
    """
    Query tool to find the origin of packages in maintained projects

    .. versionadded:: 0.3.0
    """
    _origin_priority_pattern = re.compile(
        r"^(?P<family>SUSE:[^:-]+|openSUSE:[^:]+)[:-]"
        r"(?P<major>\d+)((\.|-SP)(?P<minor>\d+))?"
        r"(:(?P<tail>.+))?$"
    )

    def __init__(self, osc_obj):
        super().__init__(osc_obj)
        self._devel_projects = None
        self._package_matrix = None

    def family_sort_key(self, key):
        """
        Provides a numeric key for sorting projects of the same family

        :param str key: Project name
        :return: Numeric value representing the version and priority of projects
        :rtype: float
        """
        if isinstance(key, int):
            return key
        if not isinstance(key, str):
            return -1

        match = self._origin_priority_pattern.match(key)
        if not match:
            return 0

        tail_mod = 0.
        if match.group("tail"):
            parts = match.group("tail").split(":")
            if "Update" in parts:
                tail_mod += .0005
            if "workarounds" in parts:
                tail_mod += .0006
            if len(parts) > 1:
                tail_mod += .0001

        value = float(match.group("major") or 0) + .01*float(match.group("minor") or 0) + tail_mod
        return value

    def family_sorter(self, unsorted_list):
        """
        Sort consecutive projects of the same family

        In case the configuration does not list the projects in a meaningful way, they are
        sorted now to avoid the need to check the package histories for every package.

        .. note::

            This sorter is only allowed to change the order of consecutive projects belonging to the
            same family.

            Consider this a workaround for
            `this issue <https://github.com/openSUSE/openSUSE-release-tools/issues/2506>`_.

        :param unsorted_list: List with project names
        :type unsorted_list: list of str
        :return: Generator with sorted list items
        :rtype: generator
        """
        i = 0
        while i < len(unsorted_list):
            match = self._origin_priority_pattern.match(unsorted_list[i])
            if not match:
                yield unsorted_list[i]
                i += 1
                continue

            end_index = -1
            for j in range(i+1, len(unsorted_list)):
                match2 = self._origin_priority_pattern.match(unsorted_list[j])
                if not match2:
                    break
                if match2.group("family") != match.group("family"):
                    break
                end_index = j
            if end_index > i:
                yield from sorted(unsorted_list[i:end_index+1], key=self.family_sort_key,
                                  reverse=True)
                i = end_index+1
            else:
                yield unsorted_list[i]
                i += 1

    @property
    def devel_projects(self):
        """
        Delayed initialization of the package specific development project dictionary

        This dictionary is populated in a lazy/JIT fashion.

        :return: :py:class:`DevelProjects`
        """
        if self._devel_projects is None:
            self._devel_projects = DevelProjects(osc_obj=self.osc)
        return self._devel_projects

    @property
    def package_matrix(self):
        """
        Delayed initialization of the lookup package matrix.

        The matrix is populated in a lazy/JIT fashion.

        :return: :py:class:`PackageMatrix`
        """
        if self._package_matrix is None:
            self._package_matrix = PackageMatrix(osc_obj=self.osc,
                                                 expanded_origins=self.expanded_origins)
        return self._package_matrix

    @cached_property
    def maintenance_project(self):
        """
        Get the maintenance project

        This project is important, because it's meta can provide the list of maintained projects.

        :return: Project name or ``None``
        :rtype: str or None

        .. versionchanged:: 0.4.0

            Return string instead of XML object
        """
        response = self.osc.search.search(path="project/id",
                                          xpath="attribute/@name='OBS:MaintenanceProject'")
        projects = getattr(response, "project", [])
        if len(projects) < 1:
            warn("The build service defines no maintenance projects!")
            return None
        if len(projects) > 1:
            warn("The build service defines multiple maintenance projects!")

        return projects[0].get("name")

    @cached_property
    def maintained_projects(self):
        """
        Get the list of maintained projects

        Maintained projects are identified by the presence of the ``OBS:Maintained`` attribute.

        :return: Project names
        :rtype: List of str

        .. versionchanged:: 0.4.0

            Search maintained projects via the ``OBS:Maintained`` attribute and not via the
            maintenance project.
        """
        response = self.osc.search.search(path="project/id",
                                          xpath="attribute/@name='OBS:Maintained'")
        return [project.get("name") for project in getattr(response, "project", [])]

    @lru_cache(maxsize=16)
    def get_project_origin_config(self, project):
        """
        Get the Origin configuration for ``project``

        This method returns the decoded YAML data, e.g. for "openSUSE:Leap:15.2:Update":

        .. code-block:: json

            {
              "origins": [
                {
                  "<devel>": {}
                },
                {
                  "SUSE:SLE-15:Update": {
                    "maintainer_review_initial": false
                  }
                },
                {
                  "SUSE:SLE-15-SP1:Update": {
                    "maintainer_review_initial": false
                  }
                },
                {
                  "SUSE:SLE-15-SP2:Update": {
                    "maintainer_review_initial": false
                  }
                },
                {
                  "openSUSE:Leap:15.1:Update": {
                    "pending_submission_allow": true
                  }
                },
                {
                  "openSUSE:Factory": {
                    "pending_submission_allow": true
                  }
                },
                {
                  "*~": {}
                }
              ],
              "fallback-group": "origin-reviewers-maintenance"
            }

        :param name project: The project name
        :return: Configuration content
        :rtype: dict
        """
        encoded_attrib = self.osc.projects.get_attribute(project=project,
                                                         attribute="OSRT:OriginConfig")

        return load(encoded_attrib.attribute.value.text, Loader=SafeLoader)

    @cached_property
    def expanded_origins(self):
        """
        Expand the projects from the Origin Configuration

        This property returns all the expanded origin projects for all projects that have the
        ``OSRT:OriginConfig`` attribute set in build service.

        .. note::

            Even if ``OSRT:OriginConfig`` explicitly defines a bogus order of projects, e.g.

            * ``SUSE:SLE-15:Update``
            * ``SUSE:SLE-15-SP1:Update``
            * ``SUSE:SLE-15-SP2:Update``

            this method will rearrange the projects in a meaningful way, e.g. into:

            * ``SUSE:SLE-15-SP2:Update``
            * ``SUSE:SLE-15-SP1:Update``
            * ``SUSE:SLE-15:Update``

            Consider this a workaround for
            `this issue <https://github.com/openSUSE/openSUSE-release-tools/issues/2506>`_.

        :return: Dictionary with lists of origin project names for all projects
        :rtype: dict
        """
        expanded = defaultdict(list)
        projects = [project.get("name")
                    for project in getattr(self.osc.search.search(
                        path="project/id", xpath="attribute/@name='OSRT:OriginConfig'"
                    ), "project", [])]

        for configured_project in projects:
            max_version = self.family_sort_key(configured_project)
            for origin in self.get_project_origin_config(configured_project).get("origins"):
                for _origin in origin:
                    _origin = _origin.rstrip("~")
                    if "*" not in _origin:
                        # no family expansion
                        expanded[configured_project].append(_origin)
                        continue

                    # family expansion
                    parts = _origin.rsplit("*", 1)
                    if parts[0] in ['', "*"]:
                        continue

                    xpath = "starts-with(@name, '{}')".format(parts[0])
                    if parts[1] != "":
                        xpath += " and contains(@name, ':{}')".format(parts[1])

                    family = [proj.get("name")
                              for proj in getattr(self.osc.search.search("project/id", xpath=xpath),
                                                  "project", [])]

                    family.sort(key=self.family_sort_key, reverse=True)
                    expanded[configured_project] += [f
                                                     for f in family
                                                     if self.family_sort_key(f) <= max_version]
            # Note: The order of origins in the config might be not the intended one. Se we sort
            # them the right way.
            expanded[configured_project] = list(self.family_sorter(expanded[configured_project]))

        return expanded

    @lru_cache(maxsize=256)
    def is_linked(self, package, project):
        """
        Guess whether a package is a link source (with incident number as suffix).

        :param package: Package name to check
        :param project: Project name containing the package
        :return: ``True``, if package name seems to have an incident ID as suffix
        """
        without_suffix = re.sub(r"\.\d+$", "", package)
        if without_suffix == package:
            return False

        package_list = {key for pkg in self.package_matrix[project].values() for key in pkg if pkg}
        if without_suffix in package_list:
            return True

        return False

    def get_origin_from_submissions(self, package, project):
        """
        Get potential origin projects from accepted submissions (aka. requests)

        This method returns the source project names from accepted requests.

        :param str package: Target package name
        :param str project: Target project name
        :return: Set of source project names
        :rtype: set or str
        """
        xpath = "action/target/@project='{project}' and action/target/@package='{package}' " \
                "and state/@name='accepted'"
        response = self.osc.search.request(xpath=xpath.format(project=project, package=package))
        sources = response.xpath("request/action/source")
        return {s.get("project") for s in sources}

    def find_package_origin(self, package, project, resolve_inheritance=True):
        """
        Find the origin project for ``package`` in ``project``

        This method will return simply ``None`` for patchinfo packages and packages with an incident
        ID as suffix.

        .. important::

            When you want to check multiple packages by multiple calls to this method, you should
            resolve from where ``project`` inherits the package for all your (package, project)
            pairs in advance and set ``resolve_inheritance`` to ``False`` to speed up the response
            time.

        :param str package: Package name
        :param str project: Project name
        :param bool resolve_inheritance: Trigger to cause this method to replace ``project`` with
                                         the project from which ``package`` is inherited from
        :return: Origin project name or ``None``
        :rtype: str
        """
        if package.startswith("patchinfo") or self.is_linked(package, project):
            # We do not care about patchinfo packages nor packages with incident ID suffixes
            return None

        if resolve_inheritance:
            response = self.osc.packages.get_files(package=package, project=project, withlinked=1)
            links = response.xpath("linkinfo/linked")
            if len(links) > 0:
                project = links[-1].get("project")

        if "<devel>" in self.expanded_origins[project] \
                and (project, package) in self.devel_projects[project]:
            return self.devel_projects[project][(project, package)]

        request_origins = self.get_origin_from_submissions(package=package, project=project) \
                          & set(self.expanded_origins[project])
        for candidate in self.expanded_origins[project]:
            if candidate in ["<devel>", "*", "*~"]:
                continue

            matrix = self.package_matrix[project][candidate]
            if package in matrix:
                origin_projects = matrix[package]
                # If the apparent origin project inherits the package, we consider the project from
                # which the package is inherited to be the correct answer.
                for proj in origin_projects[::-1]:
                    if not request_origins or proj in request_origins:
                        return origin_projects[-1]

        return None

    def list(self, project):
        """
        List all packages in a maintained project and their origin projects

        :param str project: Project name
        :return: Generator of pairs: ('pkg', 'origin')
        :rtype: generator
        """
        if project not in self.expanded_origins:
            warn("Project {} has no origin definition".format(project))
            return

        packages = self.osc.projects.get_files(project, expand='1')

        for package in getattr(packages, "entry", []):
            name = package.get("name")
            oproject = package.get("originproject")
            if name.startswith("patchinfo") or self.is_linked(name, oproject):
                # We do not care about patchinfo packages nor packages with incident ID suffixes
                continue
            yield name, self.find_package_origin(package=name, project=oproject,
                                                 resolve_inheritance=False)



"""
Distributions extension
-----------------------
"""
from __future__ import unicode_literals

from six.moves.urllib.parse import urljoin
from six import text_type

from ..utils.base import ExtensionBase


class Distribution(ExtensionBase):
    """
    Osc extension to interact with distributions

    .. versionadded:: 0.2.3
    """
    base_path = "/distributions"

    def get_list(self, include_remotes=False):
        """
        Get list of base distributions

        :param bool include_remotes: If ``True`` response will include distributions also from
                                     remote instances
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        url = urljoin(self.osc.url, self.base_path)
        if include_remotes:
            url += "/include_remotes"

        response = self.osc.request(
            url=url,
            method="GET"
        )

        return self.osc.get_objectified_xml(response)

    def get(self, distribution_id):
        """
        Get data of one base distributions hosted on this OBS instance

        :param int distribution_id: ID of distribution
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        response = self.osc.request(
            url=urljoin(
                self.osc.url,
                "/".join((self.base_path, text_type(distribution_id)))
            ),
            method="GET"
        )

        return self.osc.get_objectified_xml(response)



import re

import responses

from .base import OscTest, CallbackFactory


class OriginTest(OscTest):
    def test_origin_sort_key(self):
        pattern = self.osc.origins._origin_priority_pattern

        data = (
            # project name, expected
            ('SUSE:SLE-15:Update',
             {"family": "SUSE:SLE", "major": "15", "minor": None, "tail": "Update"}),
            ('SUSE:SLE-15-SP1:Update',
             {"family": "SUSE:SLE", "major": "15", "minor": "1", "tail": "Update"}
             ),
            ('SUSE:SLE-15-SP2:Update',
             {"family": "SUSE:SLE", "major": "15", "minor": "2", "tail": "Update"}
             ),
            ('openSUSE:Leap:15.1',
             {"family": "openSUSE:Leap", "major": "15", "minor": "1", "tail": None}
             ),
            ('openSUSE:Leap:15.1:Update',
             {"family": "openSUSE:Leap", "major": "15", "minor": "1", "tail": "Update"}
             ),
            ('openSUSE:Factory', None),
            ('openSUSE:Leap:15.1:NonFree:Update',
             {"family": "openSUSE:Leap", "major": "15", "minor": "1", "tail": "NonFree:Update"}),
            ('openSUSE:Factory:NonFree', None)
        )

        for project, expected in data:
            with self.subTest(project):
                match = pattern.match(project)
                if expected is None:
                    self.assertIsNone(match)
                else:
                    self.assertEqual(match.groupdict(), expected)

    def test_family_sorter(self):
        data = (
            # input, expected
            (
                ['<devel>', 'openSUSE:Leap:15.2:MicroOS:workarounds',
                 'SUSE:SLE-15-SP2:Update:Products:MicroOS', 'SUSE:SLE-15-SP2:Update',
                 'openSUSE:Leap:15.2:Update', 'openSUSE:Factory'],
                ['<devel>', 'openSUSE:Leap:15.2:MicroOS:workarounds',
                 'SUSE:SLE-15-SP2:Update:Products:MicroOS', 'SUSE:SLE-15-SP2:Update',
                 'openSUSE:Leap:15.2:Update', 'openSUSE:Factory']
            ),
            (
                ['<devel>', 'SUSE:SLE-15:Update', 'SUSE:SLE-15-SP1:Update',
                 'SUSE:SLE-15-SP2:Update', 'openSUSE:Leap:15.1:Update', 'openSUSE:Factory'],
                ['<devel>', 'SUSE:SLE-15-SP2:Update', 'SUSE:SLE-15-SP1:Update',
                 'SUSE:SLE-15:Update', 'openSUSE:Leap:15.1:Update', 'openSUSE:Factory']
            ),
            (
                ['<devel>', 'SUSE:SLE-15-SP2:GA', 'openSUSE:Leap:15.1:Update', 'openSUSE:Leap:15.1',
                 'openSUSE:Factory'],
                ['<devel>', 'SUSE:SLE-15-SP2:GA', 'openSUSE:Leap:15.1:Update', 'openSUSE:Leap:15.1',
                 'openSUSE:Factory']
            ),
            (
                ['<devel>', 'SUSE:SLE-15-SP2:GA', 'openSUSE:Leap:15.1', 'openSUSE:Leap:15.1:Update',
                 'openSUSE:Factory'],
                ['<devel>', 'SUSE:SLE-15-SP2:GA', 'openSUSE:Leap:15.1:Update', 'openSUSE:Leap:15.1',
                 'openSUSE:Factory']
            )
        )

        for unsorted, expected in data:
            with self.subTest():
                now_sorted = list(self.osc.origins.family_sorter(unsorted))
                self.assertTrue(len(now_sorted) == len(unsorted) == len(expected))
                self.assertEqual(expected, now_sorted)

    @responses.activate
    def test_maintained_projects(self):
        def callback(headers, params, request):
            status, body = 500, ""

            if "OBS:MaintenanceProject" in "".join(params.get("match", [])):
                status = 200
                body = """
                <collection matches="1">
                    <project name="openSUSE:Maintenance"/>
                </collection>
                """
            elif "OBS:Maintained" in "".join(params.get("match", [])):
                status = 200
                body = """
                <collection matches="2">
                    <project name="openSUSE:Leap:15.1:Update"/>
                    <project name="openSUSE:Leap:15.2:Update"/>
                </collection>
                """

            return status, headers, body

        self.mock_request(
            method=responses.GET,
            url=re.compile(self.osc.url + "/search/project/id"),
            callback=CallbackFactory(callback)
        )

        with self.subTest("Maintenance Project"):
            maint_project = self.osc.origins.maintenance_project
            self.assertEqual(maint_project, "openSUSE:Maintenance")

        with self.subTest("Maintained Projects"):
            self.assertEqual(self.osc.origins.maintained_projects,
                             ["openSUSE:Leap:15.1:Update", "openSUSE:Leap:15.2:Update"])

    @responses.activate
    def test_get_project_origin_config(self):
        self.mock_request(
            method=responses.GET,
            url=re.compile(self.osc.url + "/source/openSUSE:Leap:15.2:Update/_attribute"),
            body="""
<attributes>
  <attribute name="OriginConfig" namespace="OSRT">
    <value>origins:
- &lt;devel&gt;: {}
- SUSE:SLE-15:Update:
    maintainer_review_initial: false
- SUSE:SLE-15-SP1:Update:
    maintainer_review_initial: false
- SUSE:SLE-15-SP2:Update:
    maintainer_review_initial: false
- openSUSE:Leap:15.1:Update:
    pending_submission_allow: true
- openSUSE:Factory:
    pending_submission_allow: true
- '*~': {}
fallback-group: 'origin-reviewers-maintenance'
    </value>
  </attribute>
  <attribute name="Maintained" namespace="OBS"/>
</attributes>"""
        )

        config = self.osc.origins.get_project_origin_config("openSUSE:Leap:15.2:Update")
        self.assertEqual(config["origins"],
                         [
                             {'<devel>': {}},
                             {'SUSE:SLE-15:Update': {'maintainer_review_initial': False}},
                             {'SUSE:SLE-15-SP1:Update': {'maintainer_review_initial': False}},
                             {'SUSE:SLE-15-SP2:Update': {'maintainer_review_initial': False}},
                             {'openSUSE:Leap:15.1:Update': {'pending_submission_allow': True}},
                             {'openSUSE:Factory': {'pending_submission_allow': True}},
                             {'*~': {}}
                         ])

    @responses.activate
    def test_expanded_origins(self):
        def callback_attr(headers, params, request):
            pattern = re.compile(r"openSUSE:Leap:15\.([12]):Update")
            status, body = 500, ""
            match = pattern.search(request.url)
            if match.group(1) == "2":
                status = 200
                body = """
<attributes>
  <attribute name="OriginConfig" namespace="OSRT">
    <value>origins:
- &lt;devel&gt;: {}
- SUSE:SLE-15:Update:
    maintainer_review_initial: false
- SUSE:SLE-15-SP1:Update:
    maintainer_review_initial: false
- SUSE:SLE-15-SP2:Update:
    maintainer_review_initial: false
- openSUSE:Leap:15.1:Update:
    pending_submission_allow: true
- openSUSE:Factory:
    pending_submission_allow: true
- '*~': {}
fallback-group: 'origin-reviewers-maintenance'
    </value>
  </attribute>
  <attribute name="Maintained" namespace="OBS"/>
</attributes>"""
            elif match.group(1) == "1":
                status = 200
                body = """
<attributes>
  <attribute name="OriginConfig" namespace="OSRT">
    <value>origins:
- &lt;devel&gt;: {}
- SUSE:SLE-15*:
    maintainer_review_initial: false
- openSUSE:Factory:
    pending_submission_allow: true
- '*~': {}
fallback-group: 'origin-reviewers-maintenance'
    </value>
  </attribute>
  <attribute name="Maintained" namespace="OBS"/>
</attributes>"""

            return status, headers, body

        def callback_proj(headers, params, request):
            status, body = 500, ""
            match_string = "".join(params.get("match", []))

            if 'OSRT:OriginConfig' in match_string:
                status = 200
                body = """
                <collection matches="2">
                    <project name="openSUSE:Leap:15.1:Update"/>
                    <project name="openSUSE:Leap:15.2:Update"/>
                </collection>
                """
            elif "starts-with" in match_string:
                status = 200
                body = """
                <collection matches="10">
                  <project name='SUSE:SLE-15-SP1:GA'/>
                  <project name='SUSE:SLE-15-SP1:Update'/>
                  <project name='SUSE:SLE-15-SP2:GA'/>
                  <project name='SUSE:SLE-15-SP2:Update'/>
                  <project name='SUSE:SLE-15-SP2:Update:Products:MicroOS'/>
                  <project name='SUSE:SLE-15-SP2:Update:Products:MicroOS:Update'/>
                  <project name='SUSE:SLE-15-SP3:GA'/>
                  <project name='SUSE:SLE-15-SP3:Update'/>
                  <project name='SUSE:SLE-15:GA'/>
                  <project name='SUSE:SLE-15:Update'/>
                </collection>
                """

            return status, headers, body

        self.mock_request(
            method=responses.GET,
            url=re.compile(self.osc.url + "/search/project/id"),
            callback=CallbackFactory(callback_proj)
        )
        self.mock_request(
            method=responses.GET,
            url=re.compile(self.osc.url + r"/source/openSUSE:Leap:15\.[12]:Update/_attribute"),
            callback=CallbackFactory(callback_attr)
        )

        self.assertEqual(
            dict(self.osc.origins.expanded_origins),
            {
                "openSUSE:Leap:15.1:Update": [
                    "<devel>", "SUSE:SLE-15-SP1:Update", "SUSE:SLE-15-SP1:GA", "SUSE:SLE-15:Update",
                    "SUSE:SLE-15:GA", "openSUSE:Factory"
                ],
                "openSUSE:Leap:15.2:Update": [
                    "<devel>", "SUSE:SLE-15-SP2:Update", "SUSE:SLE-15-SP1:Update",
                    "SUSE:SLE-15:Update", "openSUSE:Leap:15.1:Update", "openSUSE:Factory"
                ]
            }
        )
